@@ -1,8 +1,7 @@
 /* gdbmdefs.h - The include file for dbm.  Defines structure and constants. */
 
 /* This file is part of GDBM, the GNU data base manager.
-   Copyright (C) 1990-1991, 1993, 2007, 2011, 2013, 2016-2020 Free
-   Software Foundation, Inc.
+   Copyright (C) 1990-2021 Free Software Foundation, Inc.
 
    GDBM is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -46,7 +45,7 @@ off_t_sum_ok (off_t a, off_t b)
 
 /* The available file space is stored in an "avail" table.  The one with
    most activity is contained in the file header. (See below.)  When that
-   one filles up, it is split in half and half is pushed on an "avail
+   one fills up, it is split in half and half is pushed on an "avail
    stack."  When the active avail table is empty and the "avail stack" is
    not empty, the top of the stack is popped into the active avail table. */
 
@@ -67,17 +66,6 @@ typedef struct
   avail_elem av_table[1]; /* The table.  Make it look like an array.  */
 } avail_block;
 
-/* Return true if avail_block is valid */
-static int inline
-gdbm_avail_block_valid_p (avail_block const *av)
-{
-  return (av->size > 1 && av->count >= 0 && av->count <= av->size);
-}
-
-/* Return true if both AV and the size of AV->av_table are valid.
-   See comment to this function in gdbmopen.c */
-extern int gdbm_avail_table_valid_p (GDBM_FILE dbf, avail_elem *av, int count);
-
 /* The dbm file header keeps track of the current location of the hash
    directory and the free space in the file.  */
 
@@ -96,7 +84,6 @@ typedef struct
                           the entire block. */
 } gdbm_file_header;
 
-
 /* The dbm hash bucket element contains the full 31 bit hash value, the
    "pointer" to the key and data (stored together) with their sizes.  It also
    has a small part of the actual key value.  It is used to verify the first
@@ -113,8 +100,6 @@ typedef struct
   int   data_size;        /* Size of associated data in the file. */
 } bucket_element;
 
-extern int gdbm_bucket_element_valid_p (GDBM_FILE dbf, int elem_loc);
-
 /* A bucket is a small hash table.  This one consists of a number of
    bucket elements plus some bookkeeping fields.  The number of elements
    depends on the optimum blocksize for the storage device and on a
@@ -157,13 +142,35 @@ typedef struct
   int     elem_loc;
 } data_cache_elem;
 
-typedef struct
+typedef struct cache_elem cache_elem;
+typedef struct cache_node cache_node;
+
+struct cache_node
+{
+  cache_node *left, *right, *parent; 
+  int color;
+  off_t adr;
+  cache_elem *elem;
+};
+
+struct cache_elem
 {
-  hash_bucket *   ca_bucket;
   off_t           ca_adr;
-  char            ca_changed;   /* Data in the bucket changed. */
-  data_cache_elem ca_data;
-} cache_elem;
+  char            ca_changed;  /* Data in the bucket changed. */
+  data_cache_elem ca_data;     /* Cached datum */
+  cache_elem      *ca_prev,    /* Previous element in LRU list. */
+                  *ca_next;    /* Next elements in LRU list.
+				  If the item is in cache_avail list, only
+				  ca_next is used.  It points to the next
+			          available element. */
+  size_t          ca_hits;     /* Number of times this element was requested */
+  cache_node      *ca_node;    /* Points back to the RBT node for this
+				  element. */
+  hash_bucket     ca_bucket[1];/* Associated  bucket (dbf->header->bucket_size
+				  bytes). */
+};
+
+typedef struct cache_tree cache_tree;
 
 /* This final structure contains all main memory based information for
    a gdbm file.  This allows multiple gdbm files to be opened at the same
@@ -199,7 +206,7 @@ struct gdbm_file_info
 
   /* Last error was fatal, the database needs recovery */
   unsigned need_recovery :1;
-  
+
   /* Last GDBM error number */
   gdbm_error last_error;
   /* Last system error number */
@@ -225,19 +232,28 @@ struct gdbm_file_info
   off_t *dir;
 
   /* The bucket cache. */
-  cache_elem *bucket_cache;
-  size_t cache_size;
-  size_t last_read;
-
+  size_t cache_size;       /* Cache capacity */
+  size_t cache_num;        /* Actual number of elements in cache */
+  /* Cache elements form a binary search tree. */
+  cache_tree *cache_tree;  
+  /* Cache elements are linked in a list sorted by relative access time */
+  cache_elem *cache_mru;   /* Most recently used element - head of the list */
+  cache_elem *cache_lru;   /* Last recently used element - tail of the list */ 
+  cache_elem *cache_avail; /* Pool of available elements (linked by prev, next)
+			    */
+  /* Pointer to the current bucket's cache entry. */
+  cache_elem *cache_entry;
+  
   /* Points to the current hash bucket in the cache. */
   hash_bucket *bucket;
-
+  
   /* The directory entry used to get the current hash bucket. */
   int bucket_dir;
 
-  /* Pointer to the current bucket's cache entry. */
-  cache_elem *cache_entry;
-
+  /* Cache statistics */
+  size_t cache_access_count; /* Number of cache accesses */
+  size_t cache_hits;         /* Number of cache hits */
+  
   /* Bookkeeping of things that need to be written back at the
      end of an update. */
   unsigned header_changed :1;
@@ -245,6 +261,9 @@ struct gdbm_file_info
   unsigned bucket_changed :1;
   unsigned second_changed :1;
 
+  off_t file_size;       /* Cached value of the current disk file size.
+			    If -1, fstat will be used to retrieve it. */
+  
   /* Mmap info */
   size_t mapped_size_max;/* Max. allowed value for mapped_size */
   void  *mapped_region;  /* Mapped region */
@@ -256,6 +275,9 @@ struct gdbm_file_info
 
 #define GDBM_DIR_COUNT(db) ((db)->header->dir_size / sizeof (off_t))
 
+#define GDBM_HEADER_AVAIL_SIZE(dbf) \
+  ((dbf)->header->block_size - offsetof (gdbm_file_header, avail))
+
 /* Execute CODE without clobbering errno */
 #define SAVE_ERRNO(code)                        \
   do                                            \
