@@ -1,8 +1,7 @@
 /* gdbmopen.c - Open the dbm file and initialize data structures for use. */
 
 /* This file is part of GDBM, the GNU data base manager.
-   Copyright (C) 1990-1991, 1993, 2007, 2011, 2013, 2016-2018 Free
-   Software Foundation, Inc.
+   Copyright (C) 1990-2021 Free Software Foundation, Inc.
 
    GDBM is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -56,82 +55,6 @@ bucket_element_count (size_t bucket_size)
   return (bucket_size - sizeof (hash_bucket)) / sizeof (bucket_element) + 1;
 }
 
-static int
-avail_comp (void const *a, void const *b)
-{
-  avail_elem const *ava = a;
-  avail_elem const *avb = b;
-  return ava->av_size - avb->av_size;
-}
-
-/* Returns true if the avail array AV[0]@COUNT is valid.
-
-   As a side effect, ensures the array is sorted by element size
-   in increasing order and restores the ordering if necessary.
-
-   The proper ordering could have been clobbered in versions of GDBM<=1.15,
-   by a call to _gdbm_put_av_elem with the can_merge parameter set to
-   TRUE. This happened in two cases: either because the GDBM_COALESCEBLKS
-   was set, and (quite unfortunately) when _gdbm_put_av_elem was called
-   from pop_avail_block in falloc.c. The latter case is quite common,
-   which means that there can be lots of existing databases with broken
-   ordering of avail arrays. Thus, restoring of the proper ordering
-   is essential for people to be able to use their existing databases.
-*/
-int
-gdbm_avail_table_valid_p (GDBM_FILE dbf, avail_elem *av, int count)
-{
-  off_t prev = 0;
-  int i;
-  int needs_sorting = 0;
-  avail_elem *p = av;
-  
-  prev = 0;
-  for (i = 0; i < count; i++, p++)
-    {
-      if (!(p->av_adr >= dbf->header->bucket_size
-	    && p->av_adr + p->av_size <= dbf->header->next_block))
-	return 0;
-      if (p->av_size < prev)
-	needs_sorting = 1;
-      prev = p->av_size;
-    }
-
-  if (needs_sorting && dbf->read_write)
-    {
-      GDBM_DEBUG (GDBM_DEBUG_ERR, "%s", "restoring sort order");
-      qsort (av, count, sizeof av[0], avail_comp);
-    }
-  
-  return 1;
-}
-
-int
-gdbm_avail_block_validate (GDBM_FILE dbf, avail_block *avblk)
-{
-  if (!(gdbm_avail_block_valid_p (avblk)
-	&& gdbm_avail_table_valid_p (dbf, avblk->av_table, avblk->count)))
-    {
-      GDBM_SET_ERRNO (dbf, GDBM_BAD_AVAIL, TRUE);
-      return -1;
-    }
-  return 0;
-}
-
-int
-gdbm_bucket_avail_table_validate (GDBM_FILE dbf, hash_bucket *bucket)
-{
-  if (!(bucket->av_count >= 0
-	&& bucket->av_count <= BUCKET_AVAIL
-	&& gdbm_avail_table_valid_p (dbf, bucket->bucket_avail,
-				     bucket->av_count)))
-    {
-      GDBM_SET_ERRNO (dbf, GDBM_BAD_AVAIL, TRUE);
-      return -1;
-    }
-  return 0;
-}
-
 static int
 validate_header (gdbm_file_header const *hdr, struct stat const *st)
 {
@@ -200,8 +123,8 @@ validate_header (gdbm_file_header const *hdr, struct stat const *st)
 
   if (((hdr->block_size - sizeof (gdbm_file_header)) / sizeof(avail_elem) + 1)
       != hdr->avail.size)
-    return GDBM_BAD_HEADER;    
-  
+    return GDBM_BAD_HEADER;
+
   return result;
 }
 
@@ -209,11 +132,19 @@ int
 _gdbm_validate_header (GDBM_FILE dbf)
 {
   struct stat file_stat;
-    
+  int rc;
+  
   if (fstat (dbf->desc, &file_stat))
     return GDBM_FILE_STAT_ERROR;
 
-  return validate_header (dbf->header, &file_stat);
+  rc = validate_header (dbf->header, &file_stat);
+  if (rc == 0)
+    {
+      if (gdbm_avail_block_validate (dbf, &dbf->header->avail,
+				     GDBM_HEADER_AVAIL_SIZE (dbf)))
+	rc = GDBM_BAD_AVAIL;
+    }
+  return rc;
 }
 
 /* Do we have ftruncate? */
@@ -269,8 +200,12 @@ gdbm_fd_open (int fd, const char *file_name, int block_size,
   dbf->dir  = NULL;
   dbf->bucket = NULL;
   dbf->header = NULL;
-  dbf->bucket_cache = NULL;
-  dbf->cache_size = 0;
+
+  /* Initialize cache */
+  dbf->cache_tree = _gdbm_cache_tree_alloc ();
+  _gdbm_cache_init (dbf, DEFAULT_CACHESIZE);
+
+  dbf->file_size = -1;
 
   dbf->memory_mapping = FALSE;
   dbf->mapped_size_max = SIZE_T_MAX;
@@ -525,7 +460,7 @@ gdbm_fd_open (int fd, const char *file_name, int block_size,
       int rc;
       
       /* Read the partial file header. */
-      if (_gdbm_full_read (dbf, &partial_header, sizeof (gdbm_file_header)))
+      if (_gdbm_full_read (dbf, &partial_header, sizeof (partial_header)))
 	{
 	  GDBM_DEBUG (GDBM_DEBUG_ERR|GDBM_DEBUG_OPEN,
 		      "%s: error reading partial header: %s",
@@ -562,27 +497,25 @@ gdbm_fd_open (int fd, const char *file_name, int block_size,
 	  return NULL;
 	}
       
-      memcpy (dbf->header, &partial_header, sizeof (gdbm_file_header));
+      memcpy (dbf->header, &partial_header, sizeof (partial_header));
       if (_gdbm_full_read (dbf, &dbf->header->avail.av_table[1],
 			   dbf->header->block_size - sizeof (gdbm_file_header)))
 	{
-	  GDBM_DEBUG (GDBM_DEBUG_ERR|GDBM_DEBUG_OPEN,
-		      "%s: error reading av_table: %s",
-		      dbf->name, gdbm_db_strerror (dbf));
 	  if (!(flags & GDBM_CLOERROR))
 	    dbf->desc = -1;
 	  SAVE_ERRNO (gdbm_close (dbf));
 	  return NULL;
 	}
 
-      if (gdbm_avail_block_validate (dbf, &dbf->header->avail))
+      if (gdbm_avail_block_validate (dbf, &dbf->header->avail,
+				     GDBM_HEADER_AVAIL_SIZE (dbf)))
 	{
 	  if (!(flags & GDBM_CLOERROR))
 	    dbf->desc = -1;
 	  SAVE_ERRNO (gdbm_close (dbf));
 	  return NULL;
-	}	  
-	
+	}
+      
       /* Allocate space for the hash table directory.  */
       dbf->dir = malloc (dbf->header->dir_size);
       if (dbf->dir == NULL)
@@ -639,15 +572,18 @@ gdbm_fd_open (int fd, const char *file_name, int block_size,
 #endif
 
   /* Finish initializing dbf. */
-  dbf->last_read = -1;
   dbf->bucket = NULL;
   dbf->bucket_dir = 0;
-  dbf->cache_entry = NULL;
   dbf->header_changed = FALSE;
   dbf->directory_changed = FALSE;
   dbf->bucket_changed = FALSE;
   dbf->second_changed = FALSE;
 
+  if (flags & GDBM_XVERIFY)
+    {
+      gdbm_avail_verify (dbf);
+    }
+  
   GDBM_DEBUG (GDBM_DEBUG_ALL, "%s: opened %s", dbf->name,
 	      dbf->need_recovery ? "for recovery" : "successfully");
 
@@ -717,46 +653,19 @@ gdbm_open (const char *file, int block_size, int flags, int mode,
 		       fatal_func);
 }
 
-/* Initialize the bucket cache. */
 int
-_gdbm_init_cache (GDBM_FILE dbf, size_t size)
+_gdbm_file_size (GDBM_FILE dbf, off_t *psize)
 {
-  int index;
-
-  if (dbf->bucket_cache == NULL)
+  if (dbf->file_size == -1)
     {
-      dbf->bucket_cache = calloc (size, sizeof(cache_elem));
-      if (dbf->bucket_cache == NULL)
-        {
-          GDBM_SET_ERRNO (dbf, GDBM_MALLOC_ERROR, TRUE);
-          return -1;
-        }
-      dbf->cache_size = size;
-
-      for (index = 0; index < size; index++)
-        {
-	  (dbf->bucket_cache[index]).ca_bucket =
-	    malloc (dbf->header->bucket_size);
-          if ((dbf->bucket_cache[index]).ca_bucket == NULL)
-	    {
-              GDBM_SET_ERRNO (dbf, GDBM_MALLOC_ERROR, TRUE);
-	      return -1;
-            }
-	  dbf->bucket_cache[index].ca_data.dptr = NULL;
-	  dbf->bucket_cache[index].ca_data.dsize = 0;
-	  _gdbm_cache_entry_invalidate (dbf, index);
-        }
-      dbf->bucket = dbf->bucket_cache[0].ca_bucket;
-      dbf->cache_entry = &dbf->bucket_cache[0];
+      struct stat sb;
+      if (fstat (dbf->desc, &sb))
+	{
+	  GDBM_SET_ERRNO (dbf, GDBM_FILE_STAT_ERROR, FALSE);
+	  return -1;
+	}
+      dbf->file_size = sb.st_size;
     }
+  *psize = dbf->file_size;
   return 0;
 }
-
-void
-_gdbm_cache_entry_invalidate (GDBM_FILE dbf, int index)
-{
-  dbf->bucket_cache[index].ca_adr = 0;
-  dbf->bucket_cache[index].ca_changed = FALSE;
-  dbf->bucket_cache[index].ca_data.hash_val = -1;
-  dbf->bucket_cache[index].ca_data.elem_loc = -1;
-}
