@@ -1,8 +1,7 @@
 /* gdbmsetopt.c - set options pertaining to a GDBM descriptor. */
 
 /* This file is part of GDBM, the GNU data base manager.
-   Copyright (C) 1993-1994, 2007, 2011, 2013, 2016-2018 Free Software
-   Foundation, Inc.
+   Copyright (C) 1993-2021 Free Software Foundation, Inc.
 
    GDBM is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -54,19 +53,13 @@ setopt_gdbm_setcachesize (GDBM_FILE dbf, void *optval, int optlen)
 {
   size_t sz;
 
-  if (dbf->bucket_cache != NULL)
-    {
-      GDBM_SET_ERRNO (dbf, GDBM_OPT_ALREADY_SET, FALSE);
-      return -1;
-    }
-  
   /* Optval will point to the new size of the cache. */
   if (get_size (optval, optlen, &sz))
     {     
       GDBM_SET_ERRNO (dbf, GDBM_OPT_ILLEGAL, FALSE);
       return -1;
     }  
-  return _gdbm_init_cache (dbf, (sz > 9) ? sz : 10);
+  return _gdbm_cache_init (dbf, sz);
 }
 
 static int
