@@ -1,8 +1,7 @@
 /* systems.h - Most of the system dependant code and defines are here. */
 
 /* This file is part of GDBM, the GNU data base manager.
-   Copyright (C) 1990-1991, 1993, 2007, 2011, 2013, 2016-2020 Free
-   Software Foundation, Inc.
+   Copyright (C) 1990-2021 Free Software Foundation, Inc.
 
    GDBM is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -20,6 +19,7 @@
 /* Include all system headers first. */
 #include <sys/types.h>
 #include <stdio.h>
+#include <stddef.h>
 #if HAVE_SYS_FILE_H
 # include <sys/file.h>
 #endif
